#include <stdio.h>
#include <stdlib.h>

#include "machinedef.h"

#ifdef VAX
#include "[.mnet]types.h"
#include "[.mnet]socket.h"
#include "[.mnet]time.h"
#include <unixio.h>
#else
#include <sys/types.h>
#include <sys/time.h>
#endif

#include <signal.h>


#include "xfer.h"
#include "readwrite.h"

extern FILE *ttyout;

/******************************************************************************
FUNCTION NAME: sigpipehandle
DESCRIPTION: Signal handler, called when the network connection (a pipe)
   is broken.
RETURNS: nothing
INPUT PARAMETERS: signal, code, scp, addr - parameters which could be used to
   reenter program.
OUTPUT PARAMETERS: none
CALLED BY: O.S.; set up in _net_read and _net_write (readwrite.c)
CALLS: exit
*/
void sigpipehandle(signal, code, scp, addr)
   int signal, code;
   struct sigcontext *scp;
   char *addr;
{
   fprintf(stderr, "Network error - remote peer closed connection.\n");
   exit(0);
} /* sigpipehandle */

/******************************************************************************
FUNCTION NAME: _net_read
DESCRIPTION: Lowest level function which retrieves data from network stream.
   Sets up function 'sigpipehandle' as a SIGPIPE signal handler.
   Uses select to generate timeout value.  Exits program if timeout occurs.
   Reads a packet of data from network, places it in buffer.
RETURNS: Return value of read or socket_read (-1 on error, size of transfer on success)
INPUT PARAMETERS: s - socket stream file descriptor
                  buf - character buffer for data
                  len - maximum number of bytes to transfer
OUTPUT PARAMETERS: none
CALLED BY: net_getchars net_readdata (readwrite.c)
CALLS: FD_ZERO, FD_SET, read (sun), socket_read (vax), select, exit
*/
int _net_read(s, buf, len)
   int s;
   char *buf;
   unsigned len;
{
   struct timeval timeout;
   struct fd_set readset;
   int sel;

   FD_ZERO(&readset);
   FD_SET(s, &readset);
   timeout.tv_sec = XFR_DATA_TO;
   timeout.tv_usec = 0;
   signal(SIGPIPE, sigpipehandle);

   do
   {
      sel = select(s + 1, &readset, (fd_set *)0, (fd_set *)0, &timeout);
      if (sel < 0)
         fprintf(stderr, "Error in select\n");
   }
   while
      (sel < 0);


   if (sel == 0)
   {
      fprintf(stderr, "Error reading from socket: network timeout (%d seconds).  Exiting.\n", 
              XFR_DATA_TO);
      exit(0);
   }


#ifdef VAX
   return(socket_read(s, buf, len));
#else
   return(read(s, buf, len));
#endif
} /* _net_read */

/******************************************************************************
FUNCTION NAME: _net_write
DESCRIPTION: Lowest level function which writes data to network stream.
   Sets up function 'sigpipehandle' as a SIGPIPE signal handler.
   Uses select to generate timeout value.  Exits program if timeout occurs.
   Writes a packet of data to network from buffer.
RETURNS: Return value of write or socket_write (-1 on error, size of 
   transfer on success)
INPUT PARAMETERS: s - socket stream file descriptor
                  buf - character buffer for data
                  len - maximum number of bytes to transfer
OUTPUT PARAMETERS: none
CALLED BY: net_putline net_putdata (readwrite.c)
CALLS: FD_ZERO, FD_SET, write (sun), socket_write (vax), select, exit
*/
int _net_write(s, buf, len)
   int s;
   char *buf;
   unsigned len;
{
   struct timeval timeout;
   struct fd_set writeset;
   int sel;
 
   FD_ZERO(&writeset);
   FD_SET(s, &writeset);
   timeout.tv_sec = XFR_DATA_TO;
   timeout.tv_usec = 0;
 
   do
   {
      sel = select(s + 1, (fd_set *)0, &writeset, (fd_set *)0, &timeout); 
      if (sel < 0)
         fprintf(stderr, "Error in select\n");
   }
   while  
      (sel < 0);
 
   if (sel == 0)
   {
      fprintf(stderr, "Error writing to socket: network timeout (%d seconds).  Exiting.\n",             
              XFR_DATA_TO);
      exit(0);
   }


#ifdef VAX
   return(socket_write(s, buf, len));
#else
   return(write(s, buf, len));
#endif
} /* _net_write */


/******************************************************************************
FUNCTION NAME: net_getchars
DESCRIPTION: Reads one character from the input socket stream, keeps track
   of the current and previous characters read.
RETURNS: Current character read in.
INPUT PARAMETERS: s - socket stream file descriptor
                  ch - pointer to current character
                  lastch - pointer to last character read in
OUTPUT PARAMETERS: ch, lastch
CALLED BY: net_getline (readwrite.c)
CALLS: _net_read (readwrite.c)
*/
int net_getchars(s, ch, lastch)
   int s;
   char *ch;
   char *lastch;
{
   static char lastchar;
   int result;

   *lastch = lastchar;
   result = _net_read(s, ch, 1);
   lastchar = *ch;

   if (result == 0)
      result = EOF;

   return(result);
} /* net_getchars */


/******************************************************************************
FUNCTION NAME: net_getline
DESCRIPTION: Reads in one line of text from a network stream
RETURNS: 1 if eof was not hit during the read, 0 if it was found
INPUT PARAMETERS: s - input socket stream file descriptor
                  line - area to store the line text.
OUTPUT PARAMETERS: none
CALLED BY: dosession (xd.c)
CALLS: net_getchars net_getline (readwrite.c)
*/
int net_getline(s, line)
   int s;
   char line[PACKETLEN];
{
   int cnt = 0;
   char ch = 0;
   char lastch = 0;
 
   /* Get current line into line buffer */
   while ((net_getchars(s, &ch, &lastch) != EOF) && (ch != CR) &&
          (ch != LF) && (cnt < PACKETLEN))
      if ((ch != LF) && (ch != CR))
         line[cnt++] = ch;
   line[cnt] = 0;
 
   /* If getline was called in the middle of a CR/LF pair, then
      get the next actual line and return */
   if ((cnt == 0) && (lastch == CR) && (ch == LF))
      return(net_getline(s, line));
 
   /* If the buffer filled up before the line ended, then throw away the
      remaining portion of the line. */
   if (cnt == PACKETLEN)
      while ((net_getchars(s, &ch, &lastch) != EOF) && (ch != CR) &&
             (ch != LF))
         {};
 
   return((ch != EOF) ? cnt : EOF);
} /* net_getline */


/******************************************************************************
FUNCTION NAME: net_putline
DESCRIPTION: Write one null-terminated line of text, of variable length, to
   the network.  strlen is used to determine string length.
RETURNS: The return value of _net_read (-1 on error, size of transfer on success)
INPUT PARAMETERS: s - socket stream
                  str - null terminated character stream to be transferred.
OUTPUT PARAMETERS: none
CALLED BY: verifyuser initfile transferfile closesession (x.c)
CALLS: _net_write (readwrite.c)
*/
int net_putline(s, str)
   int s;
   char *str;
{
   return(_net_write(s, str, strlen(str)));
} /* net_putline */

/******************************************************************************
FUNCTION NAME: net_putdata
DESCRIPTION: Write a block of data of known length to the network with a header.  
   The header to the data block gives the number of bytes to be transferred in
   this packet, followed by a colon.  The header takes 5 bytes.
RETURNS: The return value of _net_read (-1 on error, size of transfer on success)
INPUT PARAMETERS: s - socket stream
                  buf - pointer to data buffer
                  buflen - length of data block
OUTPUT PARAMETERS: none
CALLED BY: transferfile (x.c)
CALLS: _net_write (readwrite.c)
*/
int net_putdata(s, buf, buflen)
   int s;
   char *buf;
   int buflen;
{
   char header[6];

   sprintf(header, "%4d:", buflen);
   if (_net_write(s, header, 5) < 0)
      return(-1);
   else
      if (buflen != XFR_DATAFLAG)
         return(_net_write(s, buf, buflen));
      else
         return(0);
} /* net_putdata */

/******************************************************************************
FUNCTION NAME: net_readdata
DESCRIPTION: Reads a block of data, up to 'buflen' bytes, or up to the number
   of bytes specified in the packet header.  Sets 'trncnt' to the number of
   bytes expected.
RETURNS: 0 on success, -1 on error
INPUT PARAMETERS: s - socket stream
                  buf - pointer to data buffer
                  buflen - maximum number of bytes to transfer this time
OUTPUT PARAMETERS: trncnt - number of bytes actually transferred
CALLED BY: receivefile (xd.c)
CALLS: _net_read (readwrite.c)
*/
int net_readdata(s, buf, buflen, trncnt)
   int s;
   char *buf;
   int buflen;
   int *trncnt;
{
   int error;
   char header[6];
   int actualcnt;

   actualcnt = 0;
   error = 0;
   while ((error >= 0) && (actualcnt < 5))
   {
      error = _net_read(s, header + actualcnt, (5 - actualcnt));
      actualcnt += error;
   }

   if (error >= 0)
   {
      sscanf(header, "%d", trncnt);
      if (*trncnt != XFR_DATAFLAG)
      {
         if (*trncnt > buflen)
            *trncnt = buflen;
         actualcnt = 0;
         while ((error >= 0) && (actualcnt < *trncnt))
         {
            error = _net_read(s, buf + actualcnt, (*trncnt - actualcnt));
            actualcnt += error;
         }
      }
      else
         error = XFR_NOERR;

      return(error);
   }
   else
      return(-1);
} /* net_readdata */

/******************************************************************************
FUNCTION NAME: net_sendack
DESCRIPTION: Sends acknowledgement to network peer.  Ackn. begins with a 
three character number followed by a colon and a descriptive message.  This
complete message would typically be displayed for the user.  
RETURNS: Return value from _net_write (-1 on error, size of transfer on success)
INPUT PARAMETERS: s - socket stream
                  errorcode - numeric error code to be used in ackowledgement
OUTPUT PARAMETERS: none
CALLED BY: dosession main (xd.c) sprintf
CALLS: _net_write (readwrite.c)
*/
int net_sendack(s, errorcode)
   int s;
   int errorcode;
{
   char buf[PACKETLEN];
   char *msg;

   switch(errorcode)
   {
      case XFR_NOERR:     msg = XFR_NOERR_S;     break;
      case XFR_INVCMD:    msg = XFR_INVCMD_S;    break;

      case XFR_VF_NOUSER: msg = XFR_VF_NOUSER_S; break;
      case XFR_VF_NOHOME: msg = XFR_VF_NOHOME_S; break;
      case XFR_VF_TRNDIR: msg = XFR_VF_TRNDIR_S; break;
      case XFR_VF_EUID:   msg = XFR_VF_EUID_S;   break;
      case XFR_VF_MAIL:   msg = XFR_VF_MAIL_S;   break;
      case XFR_VF_UAV:    msg = XFR_VF_UAV_S;    break;
      case XFR_VF_UNV:    msg = XFR_VF_UNV_S;    break;
      case XFR_VF_ISSU:   msg = XFR_VF_ISSU_S;   break;
      case XFR_VF_WTRN:   msg = XFR_VF_WTRN_S;   break;
      case XFR_VF_OTRN:   msg = XFR_VF_OTRN_S;   break;

      case XFR_OF_COF:    msg = XFR_OF_COF_S;    break;

      case XFR_IF_COF:    msg = XFR_IF_COF_S;    break;
      case XFR_IF_DFL:    msg = XFR_IF_DFL_S;    break;
      case XFR_IF_DST:    msg = XFR_IF_DST_S;    break;
      case XFR_IF_AI:     msg = XFR_IF_AI_S;     break;

      case XFR_SD_FNI:    msg = XFR_SD_FNI_S;    break;
      case XFR_SD_CPP:    msg = XFR_SD_CPP_S;    break;
      case XFR_SD_FSZ:    msg = XFR_SD_FSZ_S;    break;

      case XFR_CS_MAIL:   msg = XFR_CS_MAIL_S;   break;

      case XFR_HLP_MSG:   msg = XFR_HLP_MSG_S;   break;

      case XFR_CSM_BCSM:  msg = XFR_CSM_BCSM_S;  break;

      default:            msg = " ";             break;
   }

   sprintf(buf, "%3d: %s\n", errorcode, msg);
   return(_net_write(s, buf, strlen(buf)));
} /* net_sendack */

/******************************************************************************
FUNCTION NAME: net_readack
DESCRIPTION: Read acknowledgement from network, return the numeric value of
   the error code to calling function.  If an error occurs, display the
   acknowledgement for the user.
RETURNS: Error code contained in the acknowledgement.
INPUT PARAMETERS: s - socket stream
OUTPUT PARAMETERS: none
CALLED BY: verifyuser initfile transferfile closesession (x.c)
CALLS: _net_getline (readwrite.c)  sscanf
*/
int net_readack(s)
   int s;
{
   char buf[PACKETLEN];
   int errorcode;

   net_getline(s, buf);
   sscanf(buf, "%d", &errorcode);
   if (errorcode != XFR_NOERR)
      fprintf(stderr, "%s\n", buf);

   return(errorcode);
} /* net_readack */


/******************************************************************************
FUNCTION NAME: getword
DESCRIPTION: Get a word specified by its location in a line of text.
RETURNS: pointer to the found word (passed in as word)
INPUT PARAMETERS: position - position of word to return
OUTPUT PARAMETERS: word - location to place found word
CALLED BY: dosession (xd.c)
CALLS: none
*/
char *getword(position, line, word)
   int position;
   char *line;
   char *word;
{
   int cnt1;
   char *ch;
   char *wordch;
 
   ch = line;
   wordch = word;
   for (cnt1 = 1; cnt1 < position; cnt1++)
   {
      while (*ch && isspace(*ch))
         ch++;
      if (*ch)
         while (*ch && !isspace(*ch))
            ch++;
   }
   if (*ch)
      while (*ch && isspace(*ch))
         ch++;
   if (*ch)
      while (*ch && !isspace(*ch))
      {
         *wordch = *ch;
         wordch++;
         ch++;

      }
   *wordch = 0;
 
   return(word);
} /* getword */

void cksum(buf, len, checksum)
   unsigned char *buf; 
   int len;
   short unsigned int *checksum;
{ 
   long unsigned int sum;
   int i = 0;
 
   sum = (long unsigned int) *checksum;

   for (; i < len; i++)
      sum += *(buf + i); 
 
   *checksum = (short unsigned int) sum % 65536;
} /* cksum */ 

