#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/vfs.h>
#include <sys/stat.h>
#include <dirent.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <fcntl.h>

#include "machinedef.h"

#include "xfer.h"

#define DEBUGVAL 0

static int DEBUG;
FILE *ttyout;

/******************************************************************************
FUNCTION NAME: verifyuser
DESCRIPTION: Verifies that the specified user exists.  Then it goes on to
   seteuid to that user (things done from that point on will be done as the
   owner of the file to be received), create a transfer directory if one does
   not exist, change the cwd to the transfer directory, and initialize a mail
   message file which will be sent to the user.
      The user is verified by reading the passwd file (either /etc/passwd or
   the NIS passwd file).  The user id number and the user's home directory is
   also found this way.  The seteuid should not fail if this program is run by
   root.
      Calling of this function is triggered by a 'USER' command from the client.
RETURNS: XFR_NOERR on success, XFR_VF_... on error
INPUT PARAMETERS: userid - name of user local to this server which will receive file
                  transfernum - id of this server process (for debug statements)
                  mailfile - address of mail file stream pointer
                  mailfilename - address of name of mail file being created
                  remotemach - name of remote (client) machine sending the file
                  remoteuser - name of remote user sending the file
OUTPUT PARAMETERS: mailfile, mailfilename
CALLED BY: dosession (xd.c)
CALLS: setpwent getpwent endpwent (<pwd>) chdir mkdir tempnam fopen fprintf (<stdio>)
*/
int verifyuser(userid, transfernum, mailfile, mailfilename, remotemach, remoteuser)
   char *userid;
   int transfernum;
   FILE **mailfile;
   char **mailfilename;
   char *remotemach, *remoteuser;
{
   int result;
   struct passwd *pw;
   int error = 0;
   int found = 0;
   struct stat fileinfo;
   int i;

   /* open passwd file, reset file pointer to beginning */
   setpwent();

   /* read from passwd file either to end or until the user is found */
   while (!error && !found)
   {
      pw = getpwent();
      if (!(error = (pw == NULL)))
         found = !strcmp(pw->pw_name, userid);
   }

   /* close the passwd file */
   endpwent();
 
 
   if (!found)
   {
      if (DEBUG)
         fprintf(ttyout, "   (%d) Error locating user %s\n", transfernum, userid);
      result = XFR_VF_NOUSER;
   }
   else if (XFR_BLOCKROOT && (pw->pw_uid == XFR_ROOTUID))
   {
      if (DEBUG)
         fprintf(ttyout, "   (%d) Attempt to transfer to privileged account - failed.\n", transfernum);
      result = XFR_VF_ISSU;
   }
   else if (seteuid(pw->pw_uid))
   {
      if (DEBUG)
         fprintf(ttyout, "   (%d) Could not set uid to %d\n", transfernum, pw->pw_uid);
      result = XFR_VF_EUID;
   }
   else if (chdir(pw->pw_dir))
   {
      if (DEBUG)
         fprintf(ttyout, "   (%d) No home directory (%s) for user %s\n", transfernum, 
                pw->pw_dir, userid);
      result = XFR_VF_NOHOME;
   }
   else if (chdir(XFR_TRNDIR))
   {
      if (mkdir(XFR_TRNDIR, XFR_TRNMODE) || chdir(XFR_TRNDIR))
      {
         if (DEBUG)
            fprintf(ttyout, "   (%d) Could not access transfer directory.\n", transfernum);
         result = XFR_VF_TRNDIR;
      }
      else
         result = XFR_NOERR;
   }
   else
   {
      result = XFR_NOERR;
   }

   if (result == XFR_NOERR)
   {
      /* Verify that the receiving user owns the transfer directory we are now in,
         and that the user has write access to it */

      stat(".", &fileinfo);
      if (!(fileinfo.st_mode & XFR_TRNWRACS))
      {
         if (DEBUG)
            fprintf(ttyout, "   (%d) User is blocking transfers (write access off)\n", 
                    transfernum);
         result = XFR_VF_WTRN;
      }
      else if (fileinfo.st_uid != geteuid())
      {
         if (DEBUG)
            fprintf(ttyout, "   (%d) User does not own transfer directory.\n",
                    transfernum);
         result = XFR_VF_OTRN;
      }
   }

   if (result == XFR_NOERR)
   {
      *mailfilename = tempnam("/tmp", "xfer_");
      if (*mailfile = fopen(*mailfilename, "w"))
      {
         fprintf(*mailfile, "Subj: FILE TRANSFER\n");
         fprintf(*mailfile, "\n");
         fprintf(*mailfile, "The following file(s) were transferred to %s's\n",
                 userid);
         fprintf(*mailfile, "transfer subdirectory from user %s on machine %s\n",
                 remoteuser, remotemach);
         fprintf(*mailfile, "\n");
      }
      else
         result = XFR_VF_MAIL;
   }
 
   if (DEBUG) 
      if (result == XFR_NOERR)
         fprintf(ttyout, "   (%d) Transferring files for user %s into dir %s/%s\n", 
                   transfernum, userid, pw->pw_dir, XFR_TRNDIR);
      else
         fprintf(ttyout, "   (%d) Error verifying userid.\n", transfernum);
              
   return(result);
} /* verifyuser */

/******************************************************************************
FUNCTION NAME: fileexists
DESCRIPTION: Check to see if a file exists by opening it and closing it.
RETURNS: 1 if the file exists, 0 if not.
INPUT PARAMETERS: filename - name of file to be checked
OUTPUT PARAMETERS: none
CALLED BY: initfile (xd.c)
CALLS: open close 
*/
int fileexists(filename, type)
   char *filename;
   int *type;
{
   DIR *dirp;
   DIR *subdir;
   struct dirent *dp;
   char linkbuf[1024];
   int exists = 0;

   dirp = opendir(".");
   for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp))
   {
      if (!strcmp(dp->d_name, filename))
      {
         exists = 1;
         if (readlink(dp->d_name, linkbuf, 1024) >= 0)
            *type = XFR_LINK;
         else if (subdir = opendir(dp->d_name))
         {
            *type = XFR_DIR;
            closedir(subdir);
         }
         else
            *type = XFR_FILE;
      }
   }
   closedir(dirp);

   return(exists);
} /* fileexists */

/******************************************************************************
FUNCTION NAME: initfile
DESCRIPTION: Check to see if the specified file will fit on its target file
   system.  If so, open the file for writing with a unique name.
RETURNS: XFR_NOERR on success, XFR_IF_... in error
INPUT PARAMETERS: f - address of output file descriptor
                  filename - name of file to be opened
                  filemod - protection mode of file to be used when file is created
                  fileblk - size of file in Sun disk blocks
OUTPUT PARAMETERS: f
CALLED BY: dosession (xd.c)
CALLS: fileexists (readwrite.c) strcpy sprintf statfs open
*/
int initfile(f, filename, filemod, fileblk)
   int *f;
   char *filename;
   long unsigned filemod;
   long unsigned fileblk;
{
   int result;
   struct statfs df;
   int diskfree;
   char tmpfilename[256];
   int filetype;
   int i;

   /* Generate a unique filename */
   strcpy(tmpfilename, filename);
   for (i = 1; fileexists(tmpfilename, &filetype) && (i < 1000); i++)
      sprintf(tmpfilename, "%s_%d", filename, i);
   strcpy(filename, tmpfilename);

   if (i < 1000)
   {
      if (statfs(".", &df) >= 0)
      {
         diskfree = (int) (100.0 * (double) (df.f_bavail - fileblk) /
                                   (double) df.f_blocks);
         if (diskfree >= XFR_DISKFREE)
            if ((*f = open(filename, O_WRONLY | O_CREAT, filemod)) >= 0)
               result = XFR_NOERR;
            else
               result = XFR_IF_COF;
         else
            result = XFR_IF_DFL;
      }
      else
         result = XFR_IF_DST;
   }
   else
      result = XFR_IF_COF;

   return(result);
} /* initfile */ 

/******************************************************************************
FUNCTION NAME: placepack
DESCRIPTION: Write to the local output file the last packet received from the
   network.
RETURNS: XFR_NOERR on success, XFR_SD_CPP on failure.
INPUT PARAMETERS: f - output file descriptor
                  packet - address of data to be written
                  packetlen - length of data packet to be written
OUTPUT PARAMETERS: none
CALLED BY: receivefile (xd.c)
CALLS: write
*/
int placepack(f, packet, packetlen)
   int f;
   char packet[];
   int packetlen;
{
   int i;
   int result;

   if (write(f, packet, packetlen) >= 0)
      result = XFR_NOERR;
   else
      result = XFR_SD_CPP;

   return(result);
} /* placepack */

/******************************************************************************
FUNCTION NAME: receivefile
DESCRIPTION: Reads 'filesiz' bytes of data from the network stream, and writes
   it to the output file.  It is assumed that the client and server will have
   already set up the file transfer so that the server will read the correct
   amount of data from the client.
RETURNS: Any error code returned from net_readdata or placepack (XFR_NOERR on success)
INPUT PARAMETERS: s - socket stream
                  f - output file descriptor
                  filesiz - size in bytes of file to be transferred.
OUTPUT PARAMETERS: none
CALLED BY: dosession (xd.c)
CALLS: net_readdata (readwrite.c) placepack (xd.c)
*/
int receivefile(s, f, filesiz, checksum)
   int s;
   int f;
   long unsigned filesiz;
   short unsigned int *checksum;
{
   long unsigned fp = 0;
   unsigned trncnt = 1;
   int error = 0;
   char buf[PACKETLEN];

   *checksum = 0;

   while ((trncnt != XFR_DATAFLAG) && (error >= 0))
   {
      error = net_readdata(s, buf, PACKETLEN, &trncnt);
      if (trncnt != XFR_DATAFLAG)
      {
         fp += trncnt;
         if (error >= 0)
         {
            cksum(buf, trncnt, checksum);
            error = placepack(f, buf, trncnt);
         }
      }
   }

/*
   if ((error == XFR_NOERR) && (filesiz != fp))
      error = XFR_SD_FSZ;
*/

   return(error);
} /* receivefile */

/******************************************************************************
FUNCTION NAME: lowercase
DESCRIPTION: Changes a string to lowercase letters.  Used to convert the
   server commands to lower case before they are parsed.
RETURNS: nothing
INPUT PARAMETERS: str - address of string to be converted
OUTPUT PARAMETERS: none
CALLED BY: dosession (xd.c)
CALLS: tolower
*/
void lowercase(str)
   char *str;
{
   int i;

   for (i = 0; *(str + i); i++)
      *(str + i) = tolower(*(str + i));
} /* lowercase */


/******************************************************************************
FUNCTION NAME: dosession
DESCRIPTION: This function is called after the server has received a connection
   from a client, and forked a subprocess.  This function handles the rest of
   the session with the client.  This function receives commands from the
   client in the form "CMND param param ...", process the command and sends back
   and acknowledgement.  After this function returns the program exits.  The 
   mainline code of this program will eventually clear out the leftover zombie
   process.
      Commands processed by this function are:
         USER serveruser clientuser - verify a local userid, setuid to it
         FILE filename filemode fileblocks filesize - initialize a file to be received
         DATA - receive data for the current file
         QUIT - quit the session, send mail message to user receiving files, exit
         HELP - send a help message back to client.  (for telnet session help)
RETURNS: nothing
INPUT PARAMETERS: s - socket stream
                  transfernum - serial number of this server session (for debugging)
                  remotemach - name of client machine (where connection came from)
OUTPUT PARAMETERS: none
CALLED BY: main (xd.c)
CALLS: lowercase verifyuser initfile receivefile (xd.c)
       getword net_sendack net_getline (readwrite.c)
       strcmp sprintf fprintf fclose close free _exit unlink 
*/
void dosession(s, transfernum, remotemach)
   int s;
   int transfernum;
   char *remotemach;
{
   char buf[PACKETLEN + 1];
   int response;
   int f;
   int done = 0;
   int error = 0;
   FILE *mailfile;
   char *mailfilename;
   char userid[256];
   char remoteuser[PACKETLEN];
   char commandstr[PACKETLEN];
   char curfilename[PACKETLEN];
   long unsigned filemod;
   long unsigned fileblk;
   long unsigned filesiz;
   long unsigned filecnt = 0;
   short unsigned int checksum;
   short unsigned int remotecksum;

   char line[PACKETLEN];
   char command[PACKETLEN];
   char word[PACKETLEN];
   int userverified = 0;
   int fileinited = 0;
   long unsigned mailresult;

   /* Continue serving until session is closed or until there is a fatal error */
   while (!done && !error)
   {
      /* Receive a command line from the network client */
      error = (net_getline(s, line) < 0);
      if (!error)
      {
         /* Find the actual command in the line, convert it to lower case */
         getword(1, line, command);
         lowercase(command);

         /* 'USER' command */
         if (!strcmp(command, XFR_CMD_USER))
         {
            if (!userverified)
            {
               getword(2, line, userid);
               getword(3, line, remoteuser);
               if (DEBUG)
                  fprintf(ttyout, "   (%d) Verifying user %s  (Remote user is %s)\n", 
                         transfernum, userid, remoteuser);
               response = verifyuser(userid, transfernum, &mailfile, &mailfilename,
                                     remotemach, remoteuser);
               userverified = (response == XFR_NOERR);
               net_sendack(s, response);
            }
            else
               net_sendack(s, XFR_VF_UAV);
         }

         /* 'FILE' command */
         else if (!strcmp(command, XFR_CMD_FILE))
         {
            if (userverified)
            {
               if (!fileinited)
               {
                  getword(2, line, curfilename);
                  filemod = atol(getword(3, line, word));
                  fileblk = atol(getword(4, line, word));
                  filesiz = atol(getword(5, line, word));
                  if (DEBUG)
                     fprintf(ttyout, "   (%d) Initializing file %s  Mode: %lu  Blocks: %lu  Size: %lu\n",
                            transfernum, curfilename, filemod, fileblk, filesiz);
      
                  response = initfile(&f, curfilename, filemod, fileblk);
                  fileinited = (response == XFR_NOERR);
                  net_sendack(s, response);
               }
               else
                  net_sendack(s, XFR_IF_AI);
            }
            else
               net_sendack(s, XFR_VF_UNV);
         }

         /* 'DATA' command */
         else if (!strcmp(command, XFR_CMD_DATA))
         {
            if (fileinited)
            {
               if (DEBUG)
                  fprintf(ttyout, "   (%d) Receiving file\n", transfernum);
               response = receivefile(s, f, filesiz, &checksum);

               if (((response == XFR_NOERR) || (response == XFR_SD_FSZ)) &&
                   mailfile)
               {
                  fprintf(mailfile, "   %s  (%lu bytes)\n", curfilename, filesiz);
                  filecnt++;
               }
               close(f);
               fileinited = 0;
               net_sendack(s, response);
            }
            else
               net_sendack(s, XFR_SD_FNI);
         }

         /* 'CKSM' command */
         else if (!strcmp(command, XFR_CMD_CKSM))
         {
            remotecksum = (short unsigned int) atol(getword(2, line, word));
            if (remotecksum == checksum)
               net_sendack(s, XFR_NOERR);
            else
            {
               if (DEBUG)
                  fprintf(ttyout, "   (%d) Checksum mismatch.  local %u remote %u\n", 
                          transfernum, checksum, remotecksum);
               net_sendack(s, XFR_CSM_BCSM);
            }
         }
 

         /* 'QUIT' command */
         else if (!strcmp(command, XFR_CMD_QUIT))
         {
            done = 1;
            if (DEBUG)
               fprintf(ttyout, "   (%d) Closing session\n", transfernum);

            /* Send mail message to user, then delete mail file */
            if (mailfile && filecnt)
            {
               fprintf(mailfile, "\nTransferred %lu file(s).\n", filecnt);
               fprintf(mailfile, ".\n");
               fclose(mailfile);

               sprintf(commandstr, "/usr/lib/sendmail %s < %s", 
                       userid, mailfilename);
               mailresult = system(commandstr);
               unlink(mailfilename);
               free(mailfilename);
            }
            if (mailresult & 0x0000ff00)
               net_sendack(s, XFR_CS_MAIL);
            else
               net_sendack(s, XFR_NOERR);
         }

         /* 'HELP' command */
         else if (!strcmp(command, XFR_CMD_HELP))
         {
            if (DEBUG)
               fprintf(ttyout, "   (%d) Sending help message to client.\n", transfernum);
            net_sendack(s, XFR_HLP_MSG);
         }

         /* no command */
         else if (command[0] == 0)
         {
            net_sendack(s, XFR_NOCMD);
         }

         /* unrecognized command */
         else
         {
            if (DEBUG)
               fprintf(ttyout, "   (%d) Unrecognized command <%s>\n", transfernum, command);
            net_sendack(s, XFR_INVCMD);
         }
      }
   }

   if ((DEBUG) && error)
       fprintf(ttyout, "error receiving packet.  Closing connection\n");

   if (DEBUG)
      fprintf(ttyout, "(%d) subprocess exiting...\n", transfernum);

   close(s);
} /* dosession */

/******************************************************************************
FUNCTION NAME: initserver
DESCRIPTION: Initialize the network session for this server.  The function
   first gets the port number to be used (unless an alternate port is specified).
   It then creates a socket, binds the socket to the port, and calls listen
   so that multiple connections can be received at one time (and be queued).
RETURNS: number of socket created.
INPUT PARAMETERS: altport - alternate port number to be optionally used
OUTPUT PARAMETERS: none
CALLED BY: main (xd.c)
CALLS: getservbyname socket bind listen
*/
int initserver(altport)
   int altport;
{
   int s;
   struct sockaddr_in sin;
   struct servent *sp;

   if (altport)
      sin.sin_port = altport;
   else
   {
      sp = (struct servent *) getservbyname(XFR_SERVICE, "tcp");
      if (sp == NULL)
      {
         if (XFR_TRYDEFSRV)
         {
            if (DEBUG)
               fprintf(ttyout, "tcp/%s: unknown service.  Attempting to use service %d\n",
                       XFR_SERVICE, XFR_DEFSRV);
            sin.sin_port = XFR_DEFSRV;
         }
         else
         {
            if (DEBUG)
               fprintf(ttyout, "tcp/%s: unknown service.\n");
            return(0);
         }
      }
      else
         sin.sin_port = sp->s_port;
   }

   sin.sin_addr.s_addr = INADDR_ANY;
   sin.sin_family = AF_INET;

   s = socket(AF_INET, SOCK_STREAM, 0);
   if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
   {
      if (DEBUG)
         fprintf(ttyout, "socket binding error\n");
      exit(1);
   }

   if (DEBUG)
      fprintf(ttyout, "Bound socket to port %d\n", sin.sin_port);

   listen(s, 5);

   return(s);
} /* initserver */


/******************************************************************************
FUNCTION NAME: clearzombies
DESCRIPTION: Clear out any existing zombie subprocesses (defunct processes:
   processes in the same process group as the main line code of this program
   which have exited)  If this function is not called periodically the system
   will fill up with defunct processes.
RETURNS: nothing
INPUT PARAMETERS: none
OUTPUT PARAMETERS: none
CALLED BY: main (xd.c)
CALLS: waitpid
*/
void clearzombies()
{
   int wpid;
   int status;

   do
   {
      wpid = waitpid(0, &status, WNOHANG);
   }
   while
      (wpid > 0);
} /* clearzombies */

/******************************************************************************
FUNCTION NAME: parseparams
DESCRIPTION: Parse command line parameters for switches -d and -p n
RETURNS: nothing
INPUT PARAMETERS: argc, argv - command line parameters
                  debug - 1 if -d option set, 0 if not
                  altport - alternate port to be used by server, taken from
                            number after -p option
OUTPUT PARAMETERS: none
CALLED BY: main (xd.c)
CALLS: strcmp atol
*/
void parseparams(argc, argv, debug, debugfile, startedbyinet)
   int argc;
   char **argv;
   int *debug;
   char **debugfile;
   int *startedbyinet;
{
   int i;

   /* Options:
         -d filename      Allow debugging text to be produced, send it to 'file'
         -i               Assume that the server was started by inet.
   */

   *debug = 0;
   *startedbyinet = 0;
   for (i = 1; *(argv + i); i++)
   {
      if (!strcmp("-d", *(argv + i)))
      {
         if (*(argv + i + 1))
         {
            *debug = 1;
            *debugfile = *(argv + i + 1);
         }
      }
      else if (!strcmp("-i", *(argv + i)))
         *startedbyinet = 1;

   }
} /* parseparams */

/******************************************************************************
FUNCTION NAME: openifaccessible
DESCRIPTION: Opens the debug file, only if it is accessible for write.  Note that
   this is required in case the debug file is a tty which is not being used 
   (writes will block to this tty if a trivial fopen is used).
RETURNS: file stream pointer of debug file if it could be opened
INPUT PARAMETERS: debugfile - name of debug file
OUTPUT PARAMETERS: none
CALLED BY: main (xd.c)
CALLS: open select fdopen FD_ZERO FD_SET
*/
FILE *openifaccessible(debugfile)
   char *debugfile;
{
   int type;
   int fd;
   FILE *fs;
   struct timeval timeout;
   struct fd_set writeset;
   int sel;
 
   FD_ZERO(&writeset);
   timeout.tv_sec = 0;
   timeout.tv_usec = 0;
   fs = 0;

   if ((fd = open(debugfile, O_WRONLY | O_CREAT | O_NDELAY, 0644)) >= 0)
   {
      FD_SET(fd, &writeset);

      sel = select(fd + 1, (fd_set *)0, &writeset, (fd_set *)0, &timeout); 
      if (sel > 0)
         fs = fdopen(fd, "a+");
   }

   return(fs);
} /* openifaccessible */


/******************************************************************************
FUNCTION NAME: main
DESCRIPTION: Main program.  Server program first sets up a socket on a well-known
   port (advertised as "xfer" in /etc/services).  Server then waits for a 
   client connect request.  As each request comes in, the server accepts a
   connection transferring it to an alternate socket, then forks a subprocess to 
   continue talking to the client.  The main line program then continues accepting
   connections.  The subprocesses exit when the connection with their client has
   terminated.
INPUT PARAMETERS: argc, argv, envp - command line parameters
CALLS: parseparams initserver (xd.c)
       net_sendack (readwrite.c)
       FD_ZERO FD_SET select accept gethostbyaddr close fork 
*/
main(argc, argv, envp)
   int argc;
   char **argv;
   char **envp;
{
   int s;
   int s2;
   struct sockaddr_in from;
   int fromlen = sizeof(from);
   fd_set ready;
   struct timeval timeout;
   int sel;
   struct hostent *hp;
   int child;
   int altport = 0;
   int startedbyinet;
   char peername[256];
   char *debugfile;

   parseparams(argc, argv, &DEBUG, &debugfile, &startedbyinet);
   DEBUG |= DEBUGVAL;

   /* Open debug file.  If it cannot be accessed, then dont print debug statements */
   if (DEBUG)
      if (!(ttyout = openifaccessible(debugfile)))
         DEBUG = 0;

   if (startedbyinet)
   {
      /* Transfer server started by inet - use standard I/O for socket stream */

      /* Get the name of the client which just connected */
      getpeername(0, (struct sockaddr *) &from, &fromlen);
      hp = gethostbyaddr((struct in_addr *)&(from.sin_addr),
                         sizeof(from.sin_addr), AF_INET);
      if (hp)
      {
         if (DEBUG)
            fprintf(ttyout, "(%d) Connection received from %s [%s]\n",
                    getpid(), hp->h_name, inet_ntoa(from.sin_addr));
         dosession(0, getpid(), hp->h_name);
      }
      else
      {  
         if (DEBUG)
            fprintf(ttyout, "(%d) Connection received from [%s]\n",
                   getpid(), inet_ntoa(from.sin_addr));
         dosession(0, getpid(), inet_ntoa(from.sin_addr));
      }

      /* The single transfer session is complete - exit. */
      exit(0);
   }
   else
   {
      /* Transfer server not started by inet - set up server socket, begin listening
         for connections.  Fork child processes to do transfers, clear out zombie processes */

      s = initserver(altport);
   
      /* Infinite loop - the server will never exit under normal conditions */
      while (s)
      {
         FD_ZERO(&ready);
         FD_SET(s, &ready);
         timeout.tv_sec = XFR_CNCT_TO;
         timeout.tv_usec = 0;
         sel = select(s + 1, &ready, (fd_set *)0, (fd_set *)0, &timeout);
   
         switch(sel)
         { 
            case -1:
               if (DEBUG)
                  fprintf(ttyout, "Error in select\n");
               break;
            case 0:
               /* Timeout:  Clear out any defunct subprocesses */
               clearzombies();
               break;
            case 1:
               /* The socket has received a request for connection */
               if (FD_ISSET(s, &ready))
               {
                  /* Accept the connection; transfer the connection to an
                     alternate socket */
                  s2 = accept(s, (struct sockaddr *) &from, &fromlen);
                  if (s2 != -1)
                  {
                     /* Get the name of the client which just connected */
                     hp = gethostbyaddr((struct in_addr *)&(from.sin_addr), 
                                        sizeof(from.sin_addr), AF_INET);
                     if (hp)
                     {
                        if (DEBUG)
                           fprintf(ttyout, "(%d) Connection received from %s [%s]\n",  
                                   getpid(), hp->h_name, inet_ntoa(from.sin_addr));
                     }
                     else
                     {
                        if (DEBUG)
                           fprintf(ttyout, "(%d) Connection received from [%s]\n", 
                                  getpid(), inet_ntoa(from.sin_addr));
                     }
   
                     /* Fork a subprocess so that the server is not blocked from
                        receiving connections.  Subprocess will finish talking to
                        client, and exit */
                     if ((child = fork()) == 0)
                     {
                        close(s);
                        if (hp)
                           dosession(s2, getpid(), hp->h_name);
                        else
                           dosession(s2, getpid(), inet_ntoa(from.sin_addr));
   
                        /* The single transfer session is complete - exit. */
                        exit(0);
                     }
                     else if (child == -1)
                     {
                        if (DEBUG)
                           fprintf(ttyout, "   (%d) Could not fork\n");
                        net_sendack(s2, XFR_IS_ERR, "could not fork");
                     }
                     close(s2);
                  }
                  else
                     if (DEBUG)
                        fprintf(ttyout, "Error in accept\n");
               }
               else
                  if (DEBUG)
                     fprintf(ttyout, "Socket not ready.  Continuing.\n");
            default:
               break;
         }
      }
   }
} /* main */
